import jax
from jax import numpy as jnp

import thermox


def test_sample_array_input():
    key = jax.random.PRNGKey(0)
    dim = 2
    dt = 0.1
    ts = jnp.arange(0, 10_000, dt)

    A = jnp.array([[3, 2], [2, 4.0]])
    b, x0 = jnp.zeros(dim), jnp.zeros(dim)
    D = 2 * jnp.eye(dim)

    samples = thermox.sample(key, ts, x0, A, b, D)

    samp_cov = jnp.cov(samples.T)
    samp_mean = jnp.mean(samples.T, axis=1)
    assert jnp.allclose(A @ samp_cov, jnp.eye(2), atol=1e-1)
    assert jnp.allclose(samp_mean, b, atol=1e-1)



from jax import numpy as jnp

from thermox.utils import (
    handle_matrix_inputs,
    ProcessedDriftMatrix,
    ProcessedDiffusionMatrix,
    preprocess,
)


def test_handle_matrix_inputs_arrays():
    A = jnp.array([[1, 3], [1, 4]])
    D = jnp.array([[9, 4], [4, 20]])

    a, d = preprocess(A, D)

    A_star, D_star = preprocess(A, D)

    assert isinstance(A_star, ProcessedDriftMatrix)
    assert isinstance(D_star, ProcessedDiffusionMatrix)
    assert jnp.all(a.val == A_star.val)


def test_handle_matrix_inputs_processed():
    A = jnp.array([[1, 3], [1, 4]])
    D = jnp.array([[9, 4], [4, 20]])

    a, d = preprocess(A, D)

    A_star, D_star = handle_matrix_inputs(a, d)

    assert isinstance(A_star, ProcessedDriftMatrix)
    assert isinstance(D_star, ProcessedDiffusionMatrix)
    assert jnp.all(a.val == A_star.val)


def test_handle_matrix_inputs_array_drift_processed_diffusion():
    A = jnp.array([[1, 3], [1, 4]])
    D = jnp.array([[9, 4], [4, 20]])

    a, d = preprocess(A, D)

    A_star, D_star = handle_matrix_inputs(A, d)

    assert isinstance(A_star, ProcessedDriftMatrix)
    assert isinstance(D_star, ProcessedDiffusionMatrix)
    assert jnp.all(a.val == A_star.val)


def test_handle_matrix_inputs_array_diffusion_processed_drift():
    A = jnp.array([[1, 3], [1, 4]])
    D = jnp.array([[9, 4], [4, 20]])

    a, d = preprocess(A, D)

    A_star, D_star = handle_matrix_inputs(a, D)

    assert isinstance(A_star, ProcessedDriftMatrix)
    assert isinstance(D_star, ProcessedDiffusionMatrix)
    assert not jnp.all(a.val == A_star.val)



import jax
from jax import numpy as jnp
import optax

import thermox


def test_log_prob_numeric_identity_diffusion():
    A_true = jnp.array([[3, 2, 1], [2, 4.0, 2], [1, 2, 5.0]])
    b_true = jnp.array([1, 2.0, 3])
    D_true = jnp.eye(3)

    nts = 1000
    ts = jnp.sort(jax.random.uniform(jax.random.PRNGKey(0), (nts,)) * 1000.0)
    x0 = jax.random.normal(jax.random.PRNGKey(0), b_true.shape)

    rk = jax.random.PRNGKey(0)
    samps = thermox.sample(rk, ts, x0, A_true, b_true, D_true)

    def transition_mean(x0, dt):
        return b_true + jax.scipy.linalg.expm(-A_true * dt) @ (x0 - b_true)

    @jax.jit
    def transition_cov_integrand(s, t):
        exp_A_st = jax.scipy.linalg.expm(A_true * (s - t))
        exp_A_T_st = jax.scipy.linalg.expm(A_true.T * (s - t))
        return exp_A_st @ D_true @ exp_A_T_st

    def transition_cov_numeric(dt):
        s_linsp = jnp.linspace(0, dt, 1000)
        evals = jax.vmap(lambda s: transition_cov_integrand(s, dt))(s_linsp)
        return jax.scipy.integrate.trapezoid(evals, s_linsp, axis=0)

    def transition_log_prob_numeric(xt, x0, t):
        mean = transition_mean(x0, t)
        cov = transition_cov_numeric(t)
        return jax.scipy.stats.multivariate_normal.logpdf(xt, mean, cov)

    log_probs_numeric = jax.vmap(
        lambda xtp1, xt, dt: transition_log_prob_numeric(xtp1, xt, dt)
    )(samps[1:], samps[:-1], ts[1:] - ts[:-1])

    log_prob = thermox.log_prob(ts, samps, A_true, b_true, D_true)

    assert jnp.isclose(log_prob, log_probs_numeric.sum(), rtol=1e-2)


def test_log_prob_numeric():
    A_true = jnp.array([[3, 2, 1], [2, 4.0, 2], [1, 2, 5.0]])
    b_true = jnp.array([1, 2.0, 3])
    D_true = jnp.array([[1, 0.3, -0.1], [0.3, 1, 0.2], [-0.1, 0.2, 1.0]])

    nts = 1000
    ts = jnp.sort(jax.random.uniform(jax.random.PRNGKey(0), (nts,)) * 1000.0)
    x0 = jax.random.normal(jax.random.PRNGKey(0), b_true.shape)

    rk = jax.random.PRNGKey(0)
    samps = thermox.sample(rk, ts, x0, A_true, b_true, D_true)

    def transition_mean(x0, dt):
        return b_true + jax.scipy.linalg.expm(-A_true * dt) @ (x0 - b_true)

    @jax.jit
    def transition_cov_integrand(s, t):
        exp_A_st = jax.scipy.linalg.expm(A_true * (s - t))
        exp_A_T_st = jax.scipy.linalg.expm(A_true.T * (s - t))
        return exp_A_st @ D_true @ exp_A_T_st

    def transition_cov_numeric(dt):
        s_linsp = jnp.linspace(0, dt, 1000)
        evals = jax.vmap(lambda s: transition_cov_integrand(s, dt))(s_linsp)
        return jax.scipy.integrate.trapezoid(evals, s_linsp, axis=0)

    def transition_log_prob_numeric(xt, x0, t):
        mean = transition_mean(x0, t)
        cov = transition_cov_numeric(t)
        return jax.scipy.stats.multivariate_normal.logpdf(xt, mean, cov)

    log_probs_numeric = jax.vmap(
        lambda xtp1, xt, dt: transition_log_prob_numeric(xtp1, xt, dt)
    )(samps[1:], samps[:-1], ts[1:] - ts[:-1])

    log_prob = thermox.log_prob(ts, samps, A_true, b_true, D_true)

    assert jnp.isclose(log_prob, log_probs_numeric.sum(), rtol=1e-2)


def test_MLE():
    A_true = jnp.array([[3, 2, 1], [2, 4.0, 2], [1, 2, 5.0]])
    b_true = jnp.array([1, 2.0, 3])
    D_true = jnp.array([[1, 0.3, -0.1], [0.3, 1, 0.2], [-0.1, 0.2, 1.0]])

    nts = 300
    ts = jnp.linspace(0, 10, nts)
    x0 = jnp.zeros_like(b_true)

    n_trajecs = 3
    rks = jax.random.split(jax.random.PRNGKey(0), n_trajecs)

    samps = jax.vmap(lambda key: thermox.sample(key, ts, x0, A_true, b_true, D_true))(
        rks
    )

    A_sqrt_init = jnp.tril(jnp.eye(3) + jax.random.normal(rks[0], (3, 3)) * 1e-1)
    b_init = jnp.zeros(3)
    D_sqrt_init = jnp.eye(3)

    log_prob_true = thermox.log_prob(ts, samps[0], A_true, b_true, D_true)
    log_prob_init = thermox.log_prob(
        ts, samps[0], A_sqrt_init @ A_sqrt_init.T, b_init, D_sqrt_init @ D_sqrt_init.T
    )

    assert log_prob_true > log_prob_init

    # Gradient descent
    def loss(params):
        A_sqrt, b, D_sqrt = params
        A_sqrt = jnp.tril(A_sqrt)
        D_sqrt = jnp.tril(D_sqrt)
        A = A_sqrt @ A_sqrt.T
        D = D_sqrt @ D_sqrt.T
        return -jax.vmap(lambda s: thermox.log_prob(ts, s, A, b, D))(
            samps
        ).mean() / len(ts)

    val_and_g = jax.jit(jax.value_and_grad(loss))

    ps = (A_sqrt_init, b_init, D_sqrt_init)
    ps_true = (jnp.linalg.cholesky(A_true), b_true, jnp.linalg.cholesky(D_true))

    v, g = val_and_g(ps)
    v_true, g_true = val_and_g(ps_true)

    assert v_true < v
    for i in range(len(ps)):
        assert jnp.all(jnp.abs(g_true[i]) <= jnp.abs(g[i]) * 1.5)

    n_steps = 20000
    neg_log_probs = jnp.zeros(n_steps)

    optimizer = optax.adam(1e-2)
    opt_state = optimizer.init(ps)

    for i in range(n_steps):
        neg_log_prob, grad = val_and_g(ps)
        if jnp.isnan(neg_log_prob) or any([jnp.isnan(g).any() for g in grad]):
            break
        updates, opt_state = optimizer.update(grad, opt_state)
        ps = optax.apply_updates(ps, updates)
        neg_log_probs = neg_log_probs.at[i].set(neg_log_prob)

    A_recover = ps[0] @ ps[0].T
    b_recover = ps[1]
    D_recover = ps[2] @ ps[2].T

    assert jnp.allclose(A_recover, A_true, atol=1e0)
    assert jnp.allclose(b_recover, b_true, atol=1e0)
    assert jnp.allclose(D_recover, D_true, atol=1e0)



from typing import NamedTuple, Tuple
from jax import numpy as jnp
from jax import Array
from fmmax.utils import (
    eig,
)  # differentiable and jit-able eigendecomposition, not yet available in jax, see https://github.com/google/jax/issues/2748


class ProcessedDriftMatrix(NamedTuple):
    """Stores eigendecompositions of A, (A+A^T)/2"""

    val: Array
    eigvals: Array
    eigvecs: Array
    eigvecs_inv: Array
    sym_eigvals: Array
    sym_eigvecs: Array


def preprocess_drift_matrix(A: Array) -> ProcessedDriftMatrix:
    """Preprocesses matrix A (calculates eigendecompositions of A and (A+A^T)/2)

    Args:
        A: Drift matrix.

    Returns:
        ProcessedDriftMatrix containing eigendeomcomposition of A and (A+A^T)/2.
    """

    A_eigvals, A_eigvecs = eig(A + 0.0j)

    A_eigvals = A_eigvals.real
    A_eigvecs = A_eigvecs.real

    A_eigvecs_inv = jnp.linalg.inv(A_eigvecs)

    symA = 0.5 * (A + A.T)
    symA_eigvals, symA_eigvecs = jnp.linalg.eigh(symA)

    return ProcessedDriftMatrix(
        A,
        A_eigvals.real,
        A_eigvecs,
        A_eigvecs_inv,
        symA_eigvals,
        symA_eigvecs,
    )


class ProcessedDiffusionMatrix(NamedTuple):
    """Stores preprocessed diffusion matrix D^0.5 and D^-0.5 via Cholesky"""

    val: Array
    sqrt: Array
    sqrt_inv: Array


def preprocess_diffusion_matrix(D: Array) -> ProcessedDiffusionMatrix:
    """Preprocesses diffusion matrix D (calculates D^0.5 and D^-0.5 via Cholesky)

    Args:
        D: Diffusion matrix.

    Returns:
        ProcessedDiffusionMatrix containing D^0.5 and D^-0.5.
    """
    D_sqrt = jnp.linalg.cholesky(D)
    D_sqrt_inv = jnp.linalg.inv(D_sqrt)
    return ProcessedDiffusionMatrix(D, D_sqrt, D_sqrt_inv)


def preprocess(
    A: Array, D: Array
) -> Tuple[ProcessedDriftMatrix, ProcessedDiffusionMatrix]:
    """Transforms the drift matrix A to A_y = D^-0.5 @ A @ D^0.5 for diffusion matrix D
    and preprocesses (calculates eigendecompositions (A_y+A_y^T)/2 as well as
    D^0.5 and D^-0.5)

    Args:
        A: Drift matrix.
        D: Diffusion matrix.

    Returns:
        ProcessedDriftMatrix containing eigendecomposition of A_y and (A_y+A_y^T)/2.
            where A_y = D^-0.5 @ A @ D^0.5
        ProcessedDiffusionMatrix containing D^0.5 and D^-0.5.
    """
    PD = preprocess_diffusion_matrix(D)
    A_y = PD.sqrt_inv @ A @ PD.sqrt
    PA_y = preprocess_drift_matrix(A_y)
    return PA_y, PD


def handle_matrix_inputs(
    A: Array | ProcessedDriftMatrix, D: Array | ProcessedDiffusionMatrix
) -> Tuple[ProcessedDriftMatrix, ProcessedDiffusionMatrix]:
    """Checks the type of the input drift matrix, A, and diffusion matrix, D,
    and ensures that they are processed in the correct way.
    Helper function for sample and log_prob functions.

    Args:
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
        D: Diffusion matrix (Array or thermox.ProcessedDiffusionMatrix).

    Returns:
        ProcessedDriftMatrix containing eigendecomposition of A_y and (A_y+A_y^T)/2.
            where A_y = D^-0.5 @ A @ D^0.5
        ProcessedDiffusionMatrix containing D^0.5 and D^-0.5.
    """
    if isinstance(A, Array) or isinstance(D, Array):
        if isinstance(A, ProcessedDriftMatrix):
            A = A.val
        if isinstance(D, ProcessedDiffusionMatrix):
            D = D.val
        A, D = preprocess(A, D)
    return A, D



import jax.numpy as jnp
from jax.lax import fori_loop
from jax import Array, vmap

from thermox.utils import (
    handle_matrix_inputs,
    preprocess_drift_matrix,
    ProcessedDriftMatrix,
    ProcessedDiffusionMatrix,
)


def log_prob_identity_diffusion(
    ts: Array,
    xs: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
) -> float:
    """Calculates log probability of samples from the Ornstein-Uhlenbeck process,
    defined as:

    dx = - A * (x - b) dt + dW

    by using exact diagonalization.

    Assumes x(t_0) is given deterministically.

    Preprocessing (diagonalisation) costs O(d^3) and evaluation then costs O(T * d^2).

    Args:
        ts: Times at which samples are collected. Includes time for x0.
        xs: Initial state of the process.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
        b: Drift displacement vector.
    Returns:
        Scalar log probability of given xs.
    """
    if isinstance(A, Array):
        A = preprocess_drift_matrix(A)

    def expm_vp(v, dt):
        out = A.eigvecs_inv @ v
        out = jnp.exp(-A.eigvals * dt) * out
        out = A.eigvecs @ out
        return out.real

    def transition_mean(y, dt):
        return b + expm_vp(y - b, dt)

    def transition_cov_sqrt_inv_vp(v, dt):
        diag = ((1 - jnp.exp(-2 * A.sym_eigvals * dt)) / (2 * A.sym_eigvals)) ** 0.5
        diag = jnp.where(diag < 1e-20, 1e-20, diag)
        out = A.sym_eigvecs.T @ v
        out = out / diag
        return out.real

    def transition_cov_log_det(dt):
        diag = (1 - jnp.exp(-2 * A.sym_eigvals * dt)) / (2 * A.sym_eigvals)
        diag = jnp.where(diag < 1e-20, 1e-20, diag)
        return jnp.sum(jnp.log(diag))

    def logpt(yt, y0, dt):
        mean = transition_mean(y0, dt)
        diff_val = transition_cov_sqrt_inv_vp(yt - mean, dt)
        return (
            -jnp.dot(diff_val, diff_val) / 2
            - transition_cov_log_det(dt) / 2
            - jnp.log(2 * jnp.pi) * (yt.shape[0] / 2)
        )

    log_prob_val = fori_loop(
        1,
        len(ts),
        lambda i, val: val + logpt(xs[i], xs[i - 1], ts[i] - ts[i - 1]),
        0.0,
    )

    return log_prob_val.real


def log_prob(
    ts: Array,
    xs: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
    D: Array | ProcessedDiffusionMatrix,
) -> Array:
    """Calculates log probability of samples from the Ornstein-Uhlenbeck process,
    defined as:

    dx = - A * (x - b) dt + sqrt(D) dW

    by using exact diagonalization.

    Assumes x(t_0) is given deterministically.

    Preprocessing (diagonalisation) costs O(d^3) and evaluation then costs O(T * d^2),
    where T=len(ts).

    By default, this function does the preprocessing on A and D before the evaluation.
    However, the preprocessing can be done externally using thermox.preprocess
    the output of which can be used as A and D here, this will skip the preprocessing.

    Args:
        ts: Times at which samples are collected. Includes time for x0.
        xs: Initial state of the process.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
            Note : If a thermox.ProcessedDriftMatrix instance is used as input,
            must be the transformed drift matrix, A_y, given by thermox.preprocess,
            not thermox.utils.preprocess_drift_matrix.
        b: Drift displacement vector.
        D: Diffusion matrix (Array or thermox.ProcessedDiffusionMatrix).

    Returns:
        Scalar log probability of given xs.
    """
    A_y, D = handle_matrix_inputs(A, D)

    ys = vmap(jnp.matmul, in_axes=(None, 0))(D.sqrt_inv, xs)
    b_y = D.sqrt_inv @ b
    log_prob_ys = log_prob_identity_diffusion(ts, ys, A_y, b_y)

    D_sqrt_inv_log_det = jnp.log(jnp.linalg.det(D.sqrt_inv))
    return log_prob_ys + D_sqrt_inv_log_det * (len(ts) - 1)



import jax
from jax import numpy as jnp

import thermox


def test_linear_system():
    A = jnp.array([[3, 2], [2, 4.0]])
    b = jnp.array([1, 2.0])

    x = thermox.linalg.solve(A, b, num_samples=10000, dt=0.1, burnin=0)

    assert jnp.allclose(A @ x, b, atol=1e-1)


def test_inv():
    A = jnp.array([[3, 2], [2, 4.0]])

    A_inv = thermox.linalg.inv(A, num_samples=10000, dt=0.1, burnin=0)

    assert jnp.allclose(A @ A_inv, jnp.eye(2), atol=1e-1)


def test_expnegm():
    A = jnp.array([[3, 2], [2, 4.0]])

    expnegA = thermox.linalg.expnegm(A, num_samples=10000, dt=0.1, burnin=0)

    assert jnp.allclose(expnegA, jax.scipy.linalg.expm(-A), atol=1e-1)


def test_expm():
    A = jnp.array([[-0.4, 0.1], [0.5, -0.3]])

    expA = thermox.linalg.expm(A, num_samples=100000, dt=0.1, burnin=0, alpha=1.0)

    assert jnp.allclose(expA, jax.scipy.linalg.expm(A), atol=1e-1)



from thermox import linalg
from thermox.sampler import sample
from thermox.prob import log_prob
from thermox.utils import preprocess
from thermox.utils import ProcessedDriftMatrix
from thermox.utils import ProcessedDiffusionMatrix



import jax
import jax.numpy as jnp
from thermox.sampler import sample, sample_identity_diffusion
from jax.lax import fori_loop
from jax import Array


def solve(
    A,
    b,
    num_samples: int = 10000,
    dt: float = 1.0,
    burnin: int = 0,
    key: Array = None,
) -> Array:
    """
    Obtain the solution of the linear system

    Ax = b

    by collecting samples from an Ornstein-Uhlenbeck
    process and calculating the mean over the samples.

    Args:
        A: Linear system matrix.
        b: Linear system vector.
        num_samples: Number of samples to be collected.
        dt: Time step.
        burnin: Time-step index corresponding to the end of the burn-in period.
            Samples before this step are not collected.
        key: JAX random key

    Returns:
        Approximate solution, x, of the linear system.
    """
    if key is None:
        key = jax.random.PRNGKey(0)
    ts = jnp.arange(burnin, burnin + num_samples) * dt
    x0 = jnp.zeros_like(b)
    samples = sample_identity_diffusion(key, ts, x0, A, jnp.linalg.solve(A, b))
    return jnp.mean(samples, axis=0)


def inv(
    A,
    num_samples: int = 10000,
    dt: float = 1.0,
    burnin: int = 0,
    key: Array = None,
) -> Array:
    """
    Obtain the inverse of a matrix A by
    collecting samples from an Ornstein-Uhlenbeck
    process and calculating the covariance of the samples.

    Args:
        A: Matrix to invert (must be symmetric positive definite).
        num_samples: Number of samples to be collected.
        dt: Time step.
        burnin: Time-step index corresponding to the end of the burn-in period.
            Samples before this step are not collected.
        key: JAX random key

    Returns:
        Approximate inverse of A.
    """
    if key is None:
        key = jax.random.PRNGKey(0)
    ts = jnp.arange(burnin, burnin + num_samples) * dt
    b = jnp.zeros(A.shape[0])
    x0 = jnp.zeros_like(b)
    samples = sample(key, ts, x0, A, b, 2 * jnp.eye(A.shape[0]))
    return jnp.cov(samples.T)


def expnegm(
    A,
    num_samples: int = 10000,
    dt: float = 1.0,
    burnin: int = 0,
    key: Array = None,
    alpha: float = 0.0,
) -> Array:
    """
    Obtain the negative exponential of a matrix A by
    collecting samples from an Ornstein-Uhlenbeck
    process and calculating the covariance of the samples.

    Args:
        A: Matrix to exponentiate.
        num_samples: Number of samples to be collected.
        dt: Time step.
        burnin: Time-step index corresponding to the end of the burn-in period.
            Samples before this step are not collected.
        key: JAX random key
        alpha: Regularization parameter to ensure diffusion matrix
            is symmetric positive definite.

    Returns:
        Approximate negative matrix exponential, exp(-A).
    """
    if key is None:
        key = jax.random.PRNGKey(0)

    A_shifted = (A + alpha * jnp.eye(A.shape[0])) / dt
    B = A_shifted + A_shifted.T

    ts = jnp.arange(burnin, burnin + num_samples) * dt
    b = jnp.zeros(A.shape[0])
    x0 = jnp.zeros_like(b)
    samples = sample(key, ts, x0, A_shifted, b, B)
    return autocovariance(samples) * jnp.exp(alpha)


def expm(
    A,
    num_samples: int = 10000,
    dt: float = 1.0,
    burnin: int = 0,
    key: Array = None,
    alpha: float = 1.0,
) -> Array:
    """
    Obtain the exponential of a matrix A by
    collecting samples from an Ornstein-Uhlenbeck
    process and calculating the covariance of the samples.

    Args:
        A: Matrix to exponentiate.
        num_samples: Number of samples to be collected.
        dt: Time step.
        burnin: Time-step index corresponding to the end of the burn-in period.
            Samples before this step are not collected.
        key: JAX random key
        alpha: Regularization parameter to ensure diffusion matrix
            is symmetric positive definite.

    Returns:
        Approximate matrix exponential, exp(A).
    """
    return expnegm(-A, num_samples, dt, burnin, key, alpha)


def autocovariance(samples: Array) -> Array:
    """
    Calculate the autocovariance of a set of samples.

    Args:
        samples: Samples from a stochastic process, as an array of shape (num_samples, dimension).

    Returns:
        Autocovariance of the samples.
    """
    return fori_loop(
        0,
        len(samples) - 1,
        lambda i, x: x + jnp.outer(samples[i + 1], samples[i]) / (len(samples) - 1),
        jnp.zeros((samples.shape[1], samples.shape[1])),
    )



import jax
import jax.numpy as jnp
from jax.lax import scan
from jax import Array

from thermox.utils import (
    handle_matrix_inputs,
    preprocess_drift_matrix,
    ProcessedDriftMatrix,
    ProcessedDiffusionMatrix,
)


def sample_identity_diffusion(
    key: Array,
    ts: Array,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
) -> Array:
    """Collects samples from the Ornstein-Uhlenbeck process, defined as:

    dx = - A * (x - b) dt + dW

    by using exact diagonalization.

    Preprocessing (diagonalisation) costs O(d^3) and sampling costs O(T * d^2)
    where T=len(ts).

    Args:
        key: Jax PRNGKey.
        ts: Times at which samples are collected. Includes time for x0.
        x0: Initial state of the process.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
        b: Drift displacement vector.

    Returns:
        Array-like, desired samples.
            shape: (len(ts), ) + x0.shape
    """

    if isinstance(A, Array):
        A = preprocess_drift_matrix(A)

    def expm_vp(v, dt):
        out = A.eigvecs_inv @ v
        out = jnp.exp(-A.eigvals * dt) * out
        out = A.eigvecs @ out
        return out.real

    def transition_mean(x, dt):
        return b + expm_vp(x - b, dt)

    def transition_cov_sqrt_vp(v, dt):
        diag = ((1 - jnp.exp(-2 * A.sym_eigvals * dt)) / (2 * A.sym_eigvals)) ** 0.5
        out = diag * v
        out = A.sym_eigvecs @ out
        return out.real

    def next_x(x, dt, tkey):
        randv = jax.random.normal(tkey, shape=x.shape)
        return transition_mean(x, dt) + transition_cov_sqrt_vp(randv, dt)

    def scan_body(x_and_key, dt):
        x, rk = x_and_key
        rk, rk_use = jax.random.split(rk)
        x = next_x(x, dt, rk_use)
        return (x, rk), x

    dts = jnp.diff(ts)

    xs = scan(scan_body, (x0, key), dts)[1]
    xs = jnp.concatenate([jnp.expand_dims(x0, axis=0), xs], axis=0)
    return xs


def sample(
    key: Array,
    ts: Array,
    x0: Array,
    A: Array | ProcessedDriftMatrix,
    b: Array,
    D: Array | ProcessedDiffusionMatrix,
) -> Array:
    """Collects samples from the Ornstein-Uhlenbeck process, defined as:

    dx = - A * (x - b) dt + sqrt(D) dW

    by using exact diagonalization.

    Preprocessing (diagonalisation) costs O(d^3) and sampling costs O(T * d^2),
    where T=len(ts).

    By default, this function does the preprocessing on A and D before the evaluation.
    However, the preprocessing can be done externally using thermox.preprocess
    the output of which can be used as A and D here, this will skip the preprocessing.

    Args:
        key: Jax PRNGKey.
        ts: Times at which samples are collected. Includes time for x0.
        x0: Initial state of the process.
        A: Drift matrix (Array or thermox.ProcessedDriftMatrix).
            Note : If a thermox.ProcessedDriftMatrix instance is used as input,
            must be the transformed drift matrix, A_y, given by thermox.preprocess,
            not thermox.utils.preprocess_drift_matrix.
        b: Drift displacement vector.
        D: Diffusion matrix (Array or thermox.ProcessedDiffusionMatrix).

    Returns:
        Array-like, desired samples.
            shape: (len(ts), ) + x0.shape
    """
    A_y, D = handle_matrix_inputs(A, D)

    y0 = D.sqrt_inv @ x0
    b_y = D.sqrt_inv @ b
    ys = sample_identity_diffusion(key, ts, y0, A_y, b_y)
    return jax.vmap(jnp.matmul, in_axes=(None, 0))(D.sqrt, ys)
